/* ====================================================================
 * Copyright (c) 2006 The httpd Garage Project or its licensors,
 * as applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied.  See the License for the specific language governing
 * permissions and limitations under the License.
 */

/*
 * Envariables and ifdef work for 1.3/2.0 compilability added by
 * Ken Coar <coar@apache.org>.
 */

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_log.h"
#include <tcpd.h>

#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)

#include "apr_strings.h"

#else

#include "ap_alloc.h"

#endif

/*
 * for libwrap
 */
#define RQ_DAEMON_NAME "httpd"
int allow_severity = LOG_INFO;
int deny_severity = LOG_WARNING;

#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)

module  AP_MODULE_DECLARE_DATA hosts_access_module;

#else

module	hosts_access_module;

#endif

typedef struct ha_config_t {
    int enabled;
#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)
    apr_table_t *envars;
#else
    table *envars;
#endif
} ha_config_t;

#define HA_OFF     0
#define HA_ON      1
#define HA_INHERIT 2

/*
 * Per-directory config record creator.
 */
#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)

static void *hosts_access_create_config(apr_pool_t *p, char *path) {
    ha_config_t	*hac;
    hac = (ha_config_t *) apr_pcalloc(p, sizeof(ha_config_t));
    hac->enabled = HA_INHERIT;
    hac->envars = apr_table_make(p, 4);
    return (void *) hac;
}

#else

static void *hosts_access_create_config(ap_pool *p, char *path) {
    ha_config_t	*hac;
    hac = (ha_config_t *) ap_pcalloc(p, sizeof(ha_config_t));
    hac->enabled = HA_INHERIT;
    hac->envars = ap_make_table(p, 4);
    return (void *) hac;
}

#endif

/*
 * Per-directory config record merge routine.
 */

#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)

static void *hosts_access_merge_config(apr_pool_t *p,
                                       void *basev, void *overridesv) {
    ha_config_t	*hac;
    ha_config_t *base = (ha_config_t *) basev;
    ha_config_t *overrides = (ha_config_t *) overridesv;

    hac = (ha_config_t *) apr_pcalloc(p, sizeof(ha_config_t));
    hac->envars = apr_table_overlay(p, base->envars, overrides->envars);
    if (overrides->enabled == HA_INHERIT) {
        hac->enabled = base->enabled;
    }
    else {
        hac->enabled = overrides->enabled;
    }
    return (void *) hac;
}

#else

static void *hosts_access_merge_config(ap_pool *p,
                                       void *basev, void *overridesv) {
    ha_config_t	*hac;
    ha_config_t *base = (ha_config_t *) basev;
    ha_config_t *overrides = (ha_config_t *) overridesv;

    hac = (ha_config_t *) ap_pcalloc(p, sizeof(ha_config_t));
    hac->envars = ap_overlay_tables(p, base->envars, overrides->envars);
    if (overrides->enabled == HA_INHERIT) {
        hac->enabled = base->enabled;
    }
    else {
        hac->enabled = overrides->enabled;
    }
    return (void *) hac;
}

#endif

/*
 * Enable/disable hosts checking in this scope.
 */
static const char *hosts_access_enable(cmd_parms *cmd, void *mconfig,
                                       int bool) {
    ha_config_t	*hac;

    hac = (ha_config_t *) mconfig;
    hac->enabled = (bool ? HA_ON : HA_OFF);

    return NULL;
}

/*
 * Mark environment variables to frob if we end up denying access.
 */
static const char *hosts_access_setenv(cmd_parms *cmd, void *mconfig,
                                       const char *arg) {
    ha_config_t *hac;
    const char *envar;
    char *enval;
    int negated;

    hac = (ha_config_t *) mconfig;
#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)
    envar = apr_pstrdup(cmd->pool, arg);
#else
    envar = ap_pstrdup(cmd->pool, arg);
#endif

    negated = (*envar == '!');
    enval = strchr(envar, '=');
    /*
     * !ENVAR=[anything] disallowed
     */
    if (negated && (enval != NULL)) {
        return "cannot negate an envariable *and* give it a value";
    }
    /*
     * ENVAR= (null string) disallowed
     */
    else if ((enval != NULL) && (*enval == '\0')) {
        return "cannot set envariable to a null value";
    }
    /*
     * !ENVAR (negation signalled by enval being set to an empty string)
     */
    else if (negated) {
        envar++;
        enval = "";
    }
    /*
     * ENVAR=n
     */
    else if (enval != NULL) {
        *enval++ = '\0';
    }
    /*
     * ENVAR (set to default value of "1"
     */
    else if (enval == NULL) {
        *enval = '\0';
        enval = "1";
    }
    /*
     * ???
     */
    else {
        return "in hac_setenv -- unhandled case!";
    }
#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)
    apr_table_setn(hac->envars, envar, enval);
#else
    ap_table_setn(hac->envars, envar, enval);
#endif
    return NULL;
}

/*
 * Per-entry routine called when running through a table.
 */

#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)

static int do_envar(void *rp, const char *key, const char *val) {
    request_rec *r = (request_rec *)rp;

    if (val == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, r,
                      "unsetting '%s'", key);
        apr_table_unset(r->subprocess_env, key);
    }
    else {
        ap_log_rerror(APLOG_MARK, LOG_DEBUG|APLOG_NOERRNO, 0, r,
                      "setting '%s' to '%s'", key, val);
        apr_table_setn(r->subprocess_env, key, val);
    }
    return 1;
}
    
#else

static int do_envar(void *rp, const char *key, const char *val) {
    request_rec *r = (request_rec *)rp;

    if (val == NULL) {
        ap_log_rerror(APLOG_MARK, LOG_DEBUG|APLOG_NOERRNO, r,
                      "unsetting '%s'", key);
        ap_table_unset(r->subprocess_env, key);
    }
    else {
        ap_log_rerror(APLOG_MARK, LOG_DEBUG|APLOG_NOERRNO, r,
                      "setting '%s' to '%s'", key, val);
        ap_table_setn(r->subprocess_env, key, val);
    }
    return 1;
}
    
#endif

/*
 * Actually check in the current scope.
 */
static int hosts_access_check_access(request_rec *r) {
    ha_config_t			*hac;
    struct request_info		ha_req;
    char			*s;

    hac = (ha_config_t *) ap_get_module_config(r->per_dir_config,
                                               &hosts_access_module);
    if (hac->enabled == HA_ON) {
        request_init(&ha_req,
                     RQ_DAEMON, RQ_DAEMON_NAME,
#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)
                     RQ_CLIENT_SIN, &r->connection->remote_addr->sa.sin,
                     RQ_SERVER_SIN, &r->connection->local_addr->sa.sin,
#else
                     RQ_CLIENT_SIN, &r->connection->remote_addr,
                     RQ_SERVER_SIN, &r->connection->local_addr,
#endif
                     NULL);
	
        s = (char *) ap_get_remote_logname(r);
        if ((s != NULL) && (*s != '\0')) {
            request_set(&ha_req, RQ_USER, s, 0);
        }
	
        sock_methods(&ha_req);
	
        if (hosts_access(&ha_req)) {
            /*
             * Any they're-allowed stuff goes here
             */
        }
        else {
#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)
            ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r,
                          "libwrap/mod_hosts_access: "
                          "connection refused from %s to %s",
                          eval_client(&ha_req), eval_server(&ha_req));
#else
            ap_log_rerror(APLOG_MARK, LOG_ERR|APLOG_NOERRNO, r,
                          "libwrap/mod_hosts_access: "
                          "connection refused from %s to %s",
                          eval_client(&ha_req), eval_server(&ha_req));
#endif
            /*
             * Now handle any envariables we're supposed to frob.
             */
#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)
            apr_table_do(do_envar, r, hac->envars, NULL);
#else
            ap_table_do(do_envar, r, hac->envars, NULL);
#endif
            return HTTP_FORBIDDEN;
        }
    }

    return DECLINED;
}

/*
 * Register hooks (2.*) and define directives (1.3 and 2.*).
 */

#if AP_MODULE_MAGIC_AT_LEAST(20020903,0)

static void hac_register_hooks(apr_pool_t *p)
{
    ap_hook_access_checker(hosts_access_check_access,
                           NULL,
                           NULL,
                           APR_HOOK_MIDDLE);
}

static const command_rec hac_cmds[] =
{
    AP_INIT_FLAG(
        "HostsAccess",
        hosts_access_enable,
        NULL,
        OR_LIMIT,
        "Check /etc/hosts.* for access"
    ),
    AP_INIT_ITERATE(
        "HostsAccessSetEnv",
        hosts_access_setenv,
        NULL,
        OR_LIMIT,
        "Environment variables to set if denying access (env[=val] ...)"
    ),
    { NULL },
};
    
module AP_MODULE_DECLARE_DATA hosts_access_module =
{
    STANDARD20_MODULE_STUFF,
    hosts_access_create_config, /* per-dir config create */
    hosts_access_merge_config,  /* per-dir merger */
    NULL,                       /* per-server config creator */
    NULL,                       /* server config merger */
    hac_cmds,                   /* command table */
    hac_register_hooks,         /* like it says */
};

#else

static const command_rec hosts_access_commands[] = {
    {
        "HostsAccess", hosts_access_enable, NULL, OR_LIMIT, FLAG,
        "On to enable hosts.allow and hosts.deny checking, Off to disable this"
    },
    {
        "HostsAccessSetEnv", hosts_access_setenv, NULL, OR_LIMIT, ITERATE,
        "Environment variables to set if denying access (env[=val] ...)"
    },
    { NULL }
};

module hosts_access_module =
{
    STANDARD_MODULE_STUFF,
    NULL,                       /* module initializer */
    hosts_access_create_config, /* per-directory config creator */
    hosts_access_merge_config,  /* dir config merger */
    NULL,                       /* server config creator */
    NULL,                       /* server config merger */
    hosts_access_commands,      /* command table */
    NULL,                       /* [9] list of handlers */
    NULL,                       /* [2] filename-to-URI translation */
    NULL,                       /* [5] check/validate user_id */
    NULL,                       /* [6] check user_id is valid *here* */
    hosts_access_check_access,  /* [4] check access by host address */
    NULL,                       /* [7] MIME type checker/setter */
    NULL,                       /* [8] fixups */
    NULL,                       /* [10] logger */
#if MODULE_MAGIC_NUMBER >= 19970103
    NULL,                       /* [3] header parser */
#endif
#if MODULE_MAGIC_NUMBER >= 19970719
    NULL,                       /* process initializer */
#endif
#if MODULE_MAGIC_NUMBER >= 19970728
    NULL,                       /* process exit/cleanup */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
    NULL                        /* [1] post read_request handling */
#endif
};

#endif
/*
 * Local Variables:
 * mode: C
 * c-file-style: "bsd"
 * tab-width: 4
 * indent-tabs-mode: nil
 * End:
 */
